from pathlib import Path
from typing import Protocol

import pandas as pd

from .typing import Format


def _store_xls(
    output: Path,
    tables: dict[str, pd.DataFrame],
) -> None:
    with pd.ExcelWriter(output) as excel_writer:
        for name, table in tables.items():
            table.to_excel(excel_writer, sheet_name=name, index=False)


def _store_md(
    output: Path,
    tables: dict[str, pd.DataFrame],
) -> None:
    output.mkdir(parents=True, exist_ok=True)

    for name, table in tables.items():
        table.to_markdown(output / f"{name}.md", index=False, tablefmt="github")


def _store_csv(
    output: Path,
    tables: dict[str, pd.DataFrame],
) -> None:
    output.mkdir(parents=True, exist_ok=True)

    for name, table in tables.items():
        table.to_csv(output / f"{name}.csv", index=False)


class _Storage(Protocol):  # pragma: no cover
    def __call__(self, output: Path, tables: dict[str, pd.DataFrame]) -> None: ...


STORAGE_DICT: dict[Format, _Storage] = {
    Format.xls: _store_xls,
    Format.md: _store_md,
    Format.csv: _store_csv,
}


def store_tables(
    format: Format,
    output: Path,
    tables: dict[str, pd.DataFrame],
) -> None:
    """Store the tables to the output directory.

    Args:
        format: The output format
        output: The output directory
        tables: The tables to store
    """
    try:
        store = STORAGE_DICT[format]
    except KeyError as e:
        raise ValueError(
            f"Invalid format '{format}'. Use 'md', 'xls', or 'csv'."
        ) from e

    store(output=output, tables=tables)



from pathlib import Path
from typing import Iterator

import pandas as pd
from pydantic import BaseModel, ConfigDict
from yaml import safe_dump, safe_load

from .random_coordinates import RandomCoordinates
from .storage import store_tables
from .typing import Format


class Coordinate(BaseModel):
    solution: int
    coordinate: str
    name: str


class TeamTable(BaseModel):
    name: str
    start_coordinate: str
    table: pd.DataFrame

    model_config = ConfigDict(arbitrary_types_allowed=True)


class Config(BaseModel):
    coordinates: list[Coordinate]
    random_coords: RandomCoordinates
    destination_coord: str

    @property
    def next_coordinates(self) -> list[Coordinate]:
        return self.coordinates[1:] + [self.coordinates[0]]

    @classmethod
    def from_yaml(cls, yaml_str: str) -> "Config":
        """Create a Config instance from a YAML string.

        Args:
            yaml_str: YAML formatted string

        Returns:
            Instance created from the YAML data
        """
        yaml_data = safe_load(yaml_str)
        return cls(**yaml_data)

    @classmethod
    def from_yaml_file(cls, yaml_file: str | Path) -> "Config":
        """Create a Config instance from a YAML file.

        Args:
            yaml_file: Path to the YAML file

        Returns:
            Instance created from the YAML file
        """
        if isinstance(yaml_file, str):
            yaml_file = Path(yaml_file)

        with open(yaml_file, "r") as f:
            return cls.from_yaml(f.read())

    def to_yaml(self) -> str:
        """Convert the model to a YAML string.

        Returns:
            The model data as a YAML formatted string
        """
        return safe_dump(self.model_dump())

    def to_yaml_file(self, yaml_file: str | Path) -> None:
        """Write the model data to a YAML file.

        Args:
            yaml_file: Path where to write the YAML file
        """
        if isinstance(yaml_file, str):
            yaml_file = Path(yaml_file)

        with open(yaml_file, "w") as f:
            f.write(self.to_yaml())

    def get_solution_next_coord_table(self) -> pd.DataFrame:
        """Get a DataFrame with the solution and the next coordinate.

        Returns:
            pd.DataFrame: DataFrame with the solution and the next coordinate
        """
        next_coords = self.next_coordinates
        return pd.DataFrame(
            {
                "solution": [coord.solution for coord in self.coordinates],
                "coordinate": [coord.coordinate for coord in next_coords],
            }
        )

    def get_overview_table(self) -> pd.DataFrame:
        """Get a DataFrame with the overview of the solutions.

        Returns:
            pd.DataFrame: DataFrame with the overview of the solutions
        """
        next_coords = self.next_coordinates
        return pd.DataFrame(
            {
                "current_name": [coord.name for coord in self.coordinates],
                "current_coordinate": [coord.coordinate for coord in self.coordinates],
                "current_solution": [coord.solution for coord in self.coordinates],
                "next_name": [coord.name for coord in next_coords],
                "next_coordinate": [coord.coordinate for coord in next_coords],
            }
        )

    def raw_player_table(self) -> pd.DataFrame:
        """Get a DataFrame with the raw player table.

        Returns:
            pd.DataFrame: DataFrame with the raw player table
        """
        correct_table = self.get_solution_next_coord_table()

        invalid_solutions = {coord.solution for coord in self.coordinates}
        random_table = self.random_coords.generate_dataframe(invalid_solutions)

        return (
            pd.concat([correct_table, random_table], ignore_index=True)
            .sort_values(by="solution")
            .loc[:, ["solution", "coordinate"]]
        )

    def team_tables(self) -> Iterator[TeamTable]:
        """Iterate over team-specific solution tables.

        Yields:
            The team tables with specific start coordinate and the common
            destination coordinate placed on the correct solution
        """
        table = self.raw_player_table()
        for coord in self.coordinates:
            table_copy = table.copy()
            table_copy.loc[table_copy["solution"] == coord.solution, "coordinate"] = (
                self.destination_coord
            )
            yield TeamTable(
                name=coord.name,
                start_coordinate=coord.coordinate,
                table=table_copy,
            )

    def save_tables(self, format: Format, output: Path) -> None:
        """Save the tables to the output directory.

        Args:
            format: The output format
            output: The output directory
        """
        tables = {
            "route": self.get_overview_table(),
        }
        starting_locations = []

        for i, team_table in enumerate(self.team_tables(), 1):
            name = f"team_{i:02d}"
            tables[name] = team_table.table
            starting_locations.append(
                {
                    "team": name,
                    "name": team_table.name,
                    "start_coordinate": team_table.start_coordinate,
                }
            )

        tables["starting_locations"] = pd.DataFrame(starting_locations)

        store_tables(
            format=format,
            output=output,
            tables=tables,
        )



from random import randint, uniform
from typing import Generic, TypeVar

import pandas as pd
from pydantic import BaseModel, model_validator

T = TypeVar("T", int, float)


class Range(BaseModel, Generic[T]):
    min: T
    max: T

    @model_validator(mode="after")
    def validate_range(self) -> "Range[T]":
        if self.min > self.max:
            raise ValueError(
                f"min value ({self.min}) must be less than or equal "
                f"to max value ({self.max})"
            )
        return self


class RandomCoordinates(BaseModel):
    count: int
    range: Range[int]
    lon: Range[float]
    lat: Range[float]

    # TODO add seed option
    def generate_dataframe(
        self, invalid_solutions: set[int] | None = None
    ) -> pd.DataFrame:
        """Generate random coordinates and append them to the existing DataFrame.

        Args:
            invalid_solutions: Set of solutions that should not be used

        Returns:
            DataFrame with random coordinates appended
        """
        invalid_solutions = invalid_solutions or set()
        result = []

        for _ in range(self.count):
            while True:
                sol = randint(self.range.min, self.range.max)
                if sol not in invalid_solutions:
                    break

            lon = uniform(self.lon.min, self.lon.max)
            lat = uniform(self.lat.min, self.lat.max)

            result.append((sol, f"{lat}, {lon}"))

        return pd.DataFrame(
            {
                "solution": [sol for sol, _ in result],
                "coordinate": [coord for _, coord in result],
            }
        )



import pytest

from waypoint_weaver.random_coordinates import RandomCoordinates, Range


def test_range_validation():
    range = Range[int](min=1, max=2)
    assert range.min == 1
    assert range.max == 2

    with pytest.raises(ValueError):
        Range[int](min=2, max=1)


def test_generate_dataframe():
    random_coords = RandomCoordinates(
        range=Range[int](min=1, max=10),
        lon=Range[float](min=1.0, max=2.0),
        lat=Range[float](min=1.0, max=2.0),
        count=3,
    )
    invalid_solutions = {1, 2}

    result_df = random_coords.generate_dataframe(invalid_solutions)
    solutions = set(result_df["solution"].values)
    coords = result_df["coordinate"].str.split(", ", expand=True).astype(float)
    lats = coords.iloc[:, 0].to_list()
    lons = coords.iloc[:, 1].to_list()

    assert len(result_df) == 3
    assert solutions.issubset(range(1, 11))
    assert not solutions.intersection(invalid_solutions)
    assert all(1.0 <= lat <= 2.0 for lat in lats)
    assert all(1.0 <= lon <= 2.0 for lon in lons)



from pathlib import Path
from tempfile import TemporaryDirectory

import pytest
from typer.testing import CliRunner

from waypoint_weaver.cli import main
from waypoint_weaver.typing import Format


@pytest.fixture
def runner() -> CliRunner:
    return CliRunner()


@pytest.mark.parametrize(
    argnames="option",
    argvalues=[
        "-o",
        "--output",
    ],
)
def test_get_config(
    runner: CliRunner,
    option: str,
):
    with TemporaryDirectory() as temp_dir:
        config = Path(temp_dir) / "config.yaml"
        result = runner.invoke(main, ["get-config", option, str(config)])
        assert result.exit_code == 0
        assert "Storing example config file at" in result.stdout
        assert config.exists()


def test_get_config_no_overwrite(
    runner: CliRunner,
):
    with TemporaryDirectory() as temp_dir:
        config = Path(temp_dir) / "config.yaml"
        config.touch()
        result = runner.invoke(
            main,
            ["get-config", "-o", str(config)],
            input="n\n",
        )
        assert result.exit_code == 1
        assert "Config file" in result.stdout
        assert "already exists. Overwrite?" in result.stdout


def test_get_config_overwrite_with_prompt(
    runner: CliRunner,
):
    with TemporaryDirectory() as temp_dir:
        config = Path(temp_dir) / "config.yaml"
        config.touch()
        result = runner.invoke(
            main,
            ["get-config", "-o", str(config)],
            input="y\n",
        )
        assert result.exit_code == 0


@pytest.mark.parametrize(
    argnames="option",
    argvalues=[
        "-y",
        "--yes",
    ],
)
def test_get_config_overwrite_with_skip(
    runner: CliRunner,
    option: str,
):
    with TemporaryDirectory() as temp_dir:
        config = Path(temp_dir) / "config.yaml"
        config.touch()
        result = runner.invoke(
            main,
            ["get-config", "-o", str(config), option],
        )
        assert result.exit_code == 0


def test_create_no_config(
    runner: CliRunner,
):
    result = runner.invoke(
        main,
        ["create"],
    )
    assert result.exit_code == 1
    assert "Config file" in result.stdout
    assert "does not exist." in result.stdout


def test_create_no_overwrite(
    runner: CliRunner,
    test_file: Path,
):
    with TemporaryDirectory() as temp_dir:
        output = Path(temp_dir) / "waypoints.xlsx"
        output.touch()
        result = runner.invoke(
            main,
            ["create", "-c", str(test_file), "-o", str(output)],
            input="n\n",
        )
        assert result.exit_code == 1
        assert "Output file" in result.stdout
        assert "already exists. Overwrite?" in result.stdout


def test_create_overwrite_with_prompt(
    runner: CliRunner,
    test_file: Path,
):
    with TemporaryDirectory() as temp_dir:
        output = Path(temp_dir) / "waypoints.xlsx"
        output.touch()
        result = runner.invoke(
            main,
            ["create", "-c", str(test_file), "-o", str(output)],
            input="y\n",
        )
        assert result.exit_code == 0


def test_create_overwrite_with_skip(
    runner: CliRunner,
    test_file: Path,
):
    with TemporaryDirectory() as temp_dir:
        output = Path(temp_dir) / "waypoints.xlsx"
        output.touch()
        result = runner.invoke(
            main,
            ["create", "-c", str(test_file), "-o", str(output), "-y"],
        )
        assert result.exit_code == 0


@pytest.mark.parametrize(
    argnames="format",
    argvalues=list(Format),
)
def test_create(
    runner: CliRunner,
    test_file: Path,
    format: Format,
):
    with TemporaryDirectory() as temp_dir:
        output = Path(temp_dir)
        if format == Format.xls:
            output = output / "waypoints.xlsx"
        else:
            output = output / "waypoints"

        result = runner.invoke(
            main,
            [
                "create",
                "--config",
                str(test_file),
                "--output",
                str(output),
                "--format",
                str(format),
            ],
        )
        assert result.exit_code == 0
        assert "Stored tables to" in result.stdout



from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Iterator

import pytest
import yaml

from waypoint_weaver.config import Config


@pytest.fixture
def n_stations() -> int:
    return 5


@pytest.fixture
def test_data(
    n_stations: int,
) -> dict[str, Any]:
    return {
        "coordinates": [
            {
                "solution": i,
                "coordinate": f"1,{i}",
                "name": f"test_{i:02d}",
            }
            for i in range(n_stations)
        ],
        "destination_coord": "1., 1.",
        "random_coords": {
            "range": {"min": 1, "max": 20},
            "lon": {"min": 1.0, "max": 1.0},
            "lat": {"min": 1.0, "max": 1.0},
            "count": 15,
        },
    }


@pytest.fixture
def test_file(test_data: dict[str, Any]) -> Iterator[Path]:
    with TemporaryDirectory() as temp_dir:
        test_file = Path(temp_dir) / "test.yaml"
        with open(test_file, "w") as f:
            yaml.safe_dump(test_data, f)
        yield test_file


@pytest.fixture
def test_config(test_data: dict[str, Any]) -> Config:
    return Config(**test_data)



from pathlib import Path
from tempfile import TemporaryDirectory
from unittest.mock import patch

import pandas as pd
import pytest

from waypoint_weaver.storage import store_tables
from waypoint_weaver.typing import Format


@pytest.fixture
def dummy_tables() -> dict[str, pd.DataFrame]:
    return {
        "table1": pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]}),
        "table2": pd.DataFrame({"c": [7, 8, 9], "d": [10, 11, 12]}),
    }


@pytest.mark.parametrize(
    argnames="format",
    argvalues=list(Format),
)
def test_store_tables(
    format: Format,
    dummy_tables: dict[str, pd.DataFrame],
):
    with TemporaryDirectory() as temp_dir:
        output = Path(temp_dir)
        if format == Format.xls:
            output = output / "test.xlsx"

        store_tables(
            format=format,
            output=output,
            tables=dummy_tables,
        )

        if format == Format.xls:
            assert output.exists()
        elif format == Format.md:
            assert (output / "table1.md").exists()
            assert (output / "table2.md").exists()
        elif format == Format.csv:
            assert (output / "table1.csv").exists()
            assert (output / "table2.csv").exists()
        else:
            raise ValueError(f"Unknown format: {format}")


@patch("waypoint_weaver.storage.STORAGE_DICT", {})
def test_empty_storage_dict():
    with pytest.raises(ValueError):
        store_tables(
            format=Format.xls,
            output=Path("test.xlsx"),
            tables={},
        )



from pathlib import Path
from typing import Annotated

from typer import Option, Typer, confirm, echo

from . import EXAMPLE_CONFIG
from .config import Config
from .typing import Format

main = Typer()


@main.command()
def get_config(
    output: Annotated[
        Path, Option("-o", "--output", help="Destination for the config file")
    ] = Path("./config.yaml"),
    skip_confirmation: Annotated[
        bool, Option("-y", "--yes", help="Skip confirmation")
    ] = False,
) -> None:
    """Get an example config file."""
    if not skip_confirmation and output.exists():
        message = f"Config file '{output}' already exists. Overwrite?"
        if not confirm(message):
            exit(1)

    echo(f"Storing example config file at {output}")

    with open(EXAMPLE_CONFIG) as f:
        text = f.read()

    with open(output, "w") as f:
        f.write(text)


@main.command()
def create(
    config_file: Annotated[
        Path, Option("-c", "--config", help="Path to config file")
    ] = Path("./config.yaml"),
    output: Annotated[
        Path,
        Option(
            "-o",
            "--output",
            help="Output file if --format=xls. Otherwise, output directory",
        ),
    ] = Path("./waypoints.xlsx"),
    format: Annotated[
        Format, Option("-f", "--format", help="Output format (md, xls, or csv)")
    ] = Format.xls,
    skip_confirmation: Annotated[
        bool, Option("-y", "--yes", help="Skip confirmation")
    ] = False,
):
    """Create one or more waypoint files from a config file."""
    if not config_file.exists():
        echo(f"Config file '{config_file}' does not exist.", err=True)
        exit(1)

    if not skip_confirmation and output.exists():
        message = f"Output file '{output}' already exists. Overwrite?"
        if not confirm(message):
            exit(1)

    config = Config.from_yaml_file(config_file)

    config.save_tables(output=output, format=format)
    echo(f"Stored tables to {output}")


if __name__ == "__main__":  # pragma: no cover
    main()



from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any
from unittest.mock import MagicMock, patch

import pytest

from waypoint_weaver.config import Config, Coordinate
from waypoint_weaver.random_coordinates import RandomCoordinates, Range
from waypoint_weaver.typing import Format


@pytest.mark.parametrize(
    argnames="cast2str",
    argvalues=[
        True,
        False,
    ],
)
def test_meta_data_from_yaml_file(
    test_file: str | Path,
    n_stations: int,
    cast2str: bool,
):
    if cast2str:
        test_file = str(test_file)

    meta_data = Config.from_yaml_file(test_file)

    assert meta_data.coordinates == [
        Coordinate(solution=i, coordinate=f"1,{i}", name=f"test_{i:02d}")
        for i in range(n_stations)
    ]
    assert meta_data.destination_coord == "1., 1."
    assert meta_data.random_coords == RandomCoordinates(
        range=Range[int](min=1, max=20),
        lon=Range[float](min=1.0, max=1.0),
        lat=Range[float](min=1.0, max=1.0),
        count=15,
    )


@pytest.mark.parametrize(
    argnames="cast2str",
    argvalues=[
        True,
        False,
    ],
)
def test_yaml_loop(
    test_data: dict[str, Any],
    cast2str: bool,
):
    meta_data = Config(**test_data)

    with TemporaryDirectory() as temp_dir:
        out: str | Path = Path(temp_dir) / "test.yaml"
        if cast2str:
            out = str(out)
        meta_data.to_yaml_file(out)
        new_meta_data = Config.from_yaml_file(out)

    assert meta_data == new_meta_data


def test_get_overview_table(
    test_config: Config,
    n_stations: int,
):
    columns = [
        "current_name",
        "current_coordinate",
        "current_solution",
        "next_name",
        "next_coordinate",
    ]

    overview_table = test_config.get_overview_table()

    assert overview_table.columns.tolist() == columns
    assert len(overview_table) == n_stations


def test_team_tables(
    test_config: Config,
    n_stations: int,
):
    team_tables = list(test_config.team_tables())

    assert len(team_tables) == n_stations
    for i, table in enumerate(team_tables):
        assert table.name == f"test_{i:02d}"
        assert table.start_coordinate == f"1,{i}"
        assert len(table.table) == test_config.random_coords.count + n_stations
        assert (
            table.table.loc[table.table["solution"] == i, "coordinate"].values[0]
            == test_config.destination_coord
        )


@pytest.mark.parametrize(
    argnames="format",
    argvalues=[
        Format.xls,
        Format.csv,
        Format.md,
    ],
)
@patch("waypoint_weaver.config.store_tables")
def test_save_tables(
    store_tables: MagicMock,
    test_config: Config,
    format: Format,
):
    output = Path("output")

    test_config.save_tables(format=format, output=output)

    store_tables.assert_called_once()
